@@ -112,12 +112,12 @@ along with
 
 /* upper and lower bounds for mifare targets uid length */
 /**
-  * Upper lenght of the uid for a valid MIFARE target
+  * Upper length of the uid for a valid MIFARE target
   */
 #define UID_LEN_UPPER 7
 
 /**
-  * Lower lenght of the uid for a valid MIFARE target
+  * Lower length of the uid for a valid MIFARE target
   */
 #define UID_LEN_LOWER 4
 
@@ -292,10 +292,9 @@ struct Product
   char *description;
 
   /**
-   * Which instance should be used for billing?
-   * Full URL, replaces the default base URL
-   * for orders involving this product.  NULL if
-   * we should use the base_url.
+   * Which instance should be used for billing?  Full URL, replaces
+   * the default base URL for orders involving this product.  NULL if
+   * we should use the #backend_base_url.
    */
   char *instance;
 
@@ -330,20 +329,25 @@ struct Product
 struct PaymentActivity
 {
   /**
-   * Handle to a PUT /proposal operation
+   * Handle to a POST /orders operation
    */
-  struct TALER_MERCHANT_ProposalOperation *po;
+  struct TALER_MERCHANT_PostOrdersOperation *po;
 
   /**
-   * Handle for a /check-payment operation.
+   * Handle for a GET /private/orders/$ID operation.
    */
-  struct TALER_MERCHANT_CheckPaymentOperation *cpo;
+  struct TALER_MERCHANT_OrderMerchantGetHandle *ogh;
 
   /**
    * The product being sold.
    */
   struct Product *product;
 
+  /**
+   * Base URL for merchant interactions for this pa.
+   */
+  const char *base_url;
+
   /**
    * Order ID for pending order
    */
@@ -540,7 +544,7 @@ struct Refund
   /**
    * Handle to the ongoing operation.
    */
-  struct TALER_MERCHANT_RefundIncreaseOperation *rio;
+  struct TALER_MERCHANT_OrderRefundHandle *orh;
 
 };
 
@@ -607,12 +611,7 @@ static char *currency;
 static char *backend_base_url;
 
 /**
- * Taler fulfillment url read from configuration file
- */
-static char *fulfillment_url;
-
-/**
- * Fulfillment message to display after successfull payment, read from configuration file
+ * Fulfillment message to display after successful payment, read from configuration file
  */
 static char *fulfillment_msg;
 
@@ -621,6 +620,11 @@ static char *fulfillment_msg;
  */
 static char *authorization;
 
+/**
+ * ESSID of a WLAN network offered by the snack system, or NULL.
+ */
+static char *essid;
+
 /**
  * Handle for the payment
  */
@@ -929,9 +933,9 @@ cleanup_payment (struct PaymentActivity *pa)
     cancelbutton_task = NULL;
   }
   if (NULL != pa->po)
-    TALER_MERCHANT_proposal_cancel (pa->po);
-  if (NULL != pa->cpo)
-    TALER_MERCHANT_check_payment_cancel (pa->cpo);
+    TALER_MERCHANT_orders_post_cancel (pa->po);
+  if (NULL != pa->ogh)
+    TALER_MERCHANT_merchant_order_get_cancel (pa->ogh);
   GNUNET_CURL_gnunet_scheduler_reschedule (&rc);
   if (NULL != pa->task)
     GNUNET_SCHEDULER_cancel (pa->task);
@@ -954,7 +958,7 @@ cleanup_payment (struct PaymentActivity *pa)
 #endif
     GNUNET_free (pa->taler_pay_uri);
   }
-  GNUNET_free_non_null (pa->order_id);
+  GNUNET_free (pa->order_id);
   GNUNET_free (pa);
 }
 
@@ -1015,7 +1019,7 @@ shutdown_task (void *cls)
     GNUNET_CONTAINER_DLL_remove (refund_head,
                                  refund_tail,
                                  r);
-    TALER_MERCHANT_refund_increase_cancel (r->rio);
+    TALER_MERCHANT_post_order_refund_cancel (r->orh);
     GNUNET_free (r);
   }
   if (NULL != context)
@@ -1108,8 +1112,8 @@ shutdown_task (void *cls)
     for (unsigned int i = 0; i < products_length; i++)
     {
       GNUNET_free (products[i].description);
-      GNUNET_free_non_null (products[i].instance);
-      GNUNET_free_non_null (products[i].preview);
+      GNUNET_free (products[i].instance);
+      GNUNET_free (products[i].preview);
     }
     GNUNET_array_grow (products,
                        products_length,
@@ -1295,7 +1299,7 @@ connect_target (void *cls)
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                 "Failed to connect to NFC target\n");
   }
-  /* if the uid lenght are out of bound abort */
+  /* if the uid length are out of bound abort */
   else if ( (pa->nt.nti.nai.szUidLen > UID_LEN_UPPER) ||
             (pa->nt.nti.nai.szUidLen < UID_LEN_LOWER) )
   {
@@ -1372,39 +1376,25 @@ start_read_keyboard (void);
  * @brief Callback to process a GET /check-payment request
  *
  * @param cls closure
- * @param http_status HTTP status code for this request
- * @param obj raw response body
- * @param paid #GNUNET_YES if the payment is settled, #GNUNET_NO if not
- *        settled, #GNUNET_SYSERR on error
- *        (note that refunded payments are returned as paid!)
- * @param refunded #GNUNET_YES if there is at least on refund on this payment,
- *        #GNUNET_NO if refunded, #GNUNET_SYSERR or error
- * @param refunded_amount amount that was refunded, NULL if there
- *        was no refund
- * @param taler_pay_uri the URI that instructs the wallets to process
- *                      the payment
+ * @param hr HTTP response
+ * @param osr order status response details (on success)
  */
 static void
 check_payment_cb (void *cls,
-                  unsigned int http_status,
-                  const json_t *obj,
-                  int paid,
-                  int refunded,
-                  struct TALER_Amount *refund_amount,
-                  const char *taler_pay_uri)
+                  const struct TALER_MERCHANT_HttpResponse *hr,
+                  const struct TALER_MERCHANT_OrderStatusResponse *osr)
 {
   struct PaymentActivity *pa = cls;
+  char *uri;
 
-  (void) refunded;
-  (void) refund_amount;
-  (void) obj;
   GNUNET_assert (payment_activity == pa);
-  pa->cpo = NULL;
-  if (MHD_HTTP_OK != http_status)
+  pa->ogh = NULL;
+  if (MHD_HTTP_OK != hr->http_status)
   {
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                "Backend request to /check-payment failed: %u\n",
-                http_status);
+                "Backend request to /check-payment failed: %u/%d\n",
+                hr->http_status,
+                (int) hr->ec);
     mdb.cmd = &cmd_reader_display_backend_not_reachable;
     run_mdb_event_loop ();
     cleanup_payment (pa);
@@ -1413,7 +1403,7 @@ check_payment_cb (void *cls,
     return;
   }
 
-  if (paid)
+  if (TALER_MERCHANT_OSC_PAID == osr->status)
   {
     mdb.cmd = &cmd_approve_vend;
     payment_activity->paid = GNUNET_YES;
@@ -1426,29 +1416,34 @@ check_payment_cb (void *cls,
     }
     return;
   }
-  else
+  /* Start to check for payment. Note that we do this even before
+     we talked successfully to the wallet via NFC because we MAY show the
+     QR code in the future and in that case the payment may happen
+     anytime even before the NFC communication succeeds. */
+  if ( (NULL == pa->ogh) &&
+       (NULL == pa->delay_pay_task) )
   {
-    /* Start to check for payment. Note that we do this even before
-       we talked successfully to the wallet via NFC because we MAY show the
-       QR code in the future and in that case the payment may happen
-       anytime even before the NFC communication succeeds. */
-    if ( (NULL == pa->cpo) &&
-         (NULL == pa->delay_pay_task) )
-    {
-      pa->delay_pay_task = GNUNET_SCHEDULER_add_delayed (MAX_HTTP_RETRY_FREQ,
-                                                         &check_payment_again,
-                                                         pa);
-    }
+    pa->delay_pay_task = GNUNET_SCHEDULER_add_delayed (MAX_HTTP_RETRY_FREQ,
+                                                       &check_payment_again,
+                                                       pa);
   }
-  if (NULL == pa->taler_pay_uri)
+  if ( (NULL == pa->taler_pay_uri) &&
+       (TALER_MERCHANT_OSC_UNPAID == osr->status) )
   {
+    if (NULL == essid)
+      uri = GNUNET_strdup (osr->details.unpaid.taler_pay_uri);
+    else
+      GNUNET_asprintf (&uri,
+                       "%s#%s",
+                       osr->details.unpaid.taler_pay_uri,
+                       essid);
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                 "Trying to talk to wallet to give it pay URI `%s'\n",
-                taler_pay_uri);
+                uri);
     GNUNET_assert (NULL == pa->pnd);
-    pa->taler_pay_uri = GNUNET_strdup (taler_pay_uri);
+    pa->taler_pay_uri = uri;
 #if HAVE_QRENCODE_H
-    show_qrcode (taler_pay_uri);
+    show_qrcode (uri);
 #endif
     pa->task = GNUNET_SCHEDULER_add_now (&open_nfc_reader,
                                          pa);
@@ -1465,49 +1460,47 @@ static void
 check_payment_again (void *cls)
 {
   struct PaymentActivity *pa = cls;
-  struct Product *p = pa->product;
 
   pa->delay_pay_task = NULL;
-  GNUNET_assert (NULL == pa->cpo);
-  pa->cpo = TALER_MERCHANT_check_payment (ctx,
-                                          (NULL == p->instance)
-                                          ? backend_base_url
-                                          : p->instance,
-                                          pa->order_id,
-                                          NULL /* snack machine, no Web crap */,
-                                          BACKEND_POLL_TIMEOUT,
-                                          &check_payment_cb,
-                                          pa);
+  GNUNET_assert (NULL == pa->ogh);
+  pa->ogh = TALER_MERCHANT_merchant_order_get (ctx,
+                                               pa->base_url,
+                                               pa->order_id,
+                                               NULL /* snack machine, no Web crap */,
+                                               false,
+                                               BACKEND_POLL_TIMEOUT,
+                                               &check_payment_cb,
+                                               pa);
 }
 
 
 /**
- * @brief Callback for a PUT /order request
+ * @brief Callback for a POST /orders request
  *
  * @param cls closure
- * @param http_status HTTP status code for this request
- * @param ec Taler error code
- * @param obj raw response body
+ * @param hr HTTP response for this request
  * @param order_id order ID of the order created
+ * @param token the claim token generated by the merchant (NULL if
+ *        it wasn't generated).
  */
 static void
 proposal_cb (void *cls,
-             unsigned int http_status,
-             enum TALER_ErrorCode ec,
-             const json_t *obj,
-             const char *order_id)
+             const struct TALER_MERCHANT_HttpResponse *hr,
+             const char *order_id,
+             const struct TALER_ClaimTokenP *token)
 {
-  (void) obj;
   struct PaymentActivity *pa = cls;
 
   pa->po = NULL;
-  if (MHD_HTTP_OK != http_status)
+  if (MHD_HTTP_OK != hr->http_status)
   {
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                 "Failed to setup order with backend: %u/%d\n",
-                http_status,
-                (int) ec);
-    json_dumpf (obj, stderr, 0);
+                hr->http_status,
+                (int) hr->ec);
+    json_dumpf (hr->reply,
+                stderr,
+                0);
     mdb.cmd = &cmd_reader_display_backend_not_reachable;
     run_mdb_event_loop ();
     cleanup_payment (pa);
@@ -1519,13 +1512,14 @@ proposal_cb (void *cls,
               "Backend successfully created order `%s'\n",
               order_id);
   pa->order_id = GNUNET_strdup (order_id);
-  pa->cpo = TALER_MERCHANT_check_payment (ctx,
-                                          backend_base_url,
-                                          pa->order_id,
-                                          NULL /* snack machine, no Web crap */,
-                                          GNUNET_TIME_UNIT_ZERO,
-                                          &check_payment_cb,
-                                          pa);
+  pa->ogh = TALER_MERCHANT_merchant_order_get (ctx,
+                                               pa->base_url,
+                                               pa->order_id,
+                                               NULL /* snack machine, no Web crap */,
+                                               false,
+                                               GNUNET_TIME_UNIT_ZERO,
+                                               &check_payment_cb,
+                                               pa);
 }
 
 
@@ -1543,51 +1537,59 @@ launch_payment (struct Product *product)
 {
   struct PaymentActivity *pa;
   json_t *orderReq;
-  char *fulflmntUrl;
-  struct GNUNET_ShortHashCode uuid;
-  char *uuid_s;
+  char *msg;
+  const char *pos;
 
-  /* We need to ensure that every fulfillment URL is unique;
-     most easily done by adding a random nonce, as we may
-     not have a reliable counter or clock on the machine */
-  GNUNET_CRYPTO_random_block (GNUNET_CRYPTO_QUALITY_NONCE,
-                              &uuid,
-                              sizeof (uuid));
-  uuid_s = GNUNET_STRINGS_data_to_string_alloc (&uuid,
-                                                sizeof (uuid));
-  /* create the fulfillment url, e.g. "taler://fulfillment-success/Enjoy+your+ice+cream!"; */
-  GNUNET_asprintf (&fulflmntUrl,
-                   "%s%s%s#%s",
-                   fulfillment_url,
-                   fulfillment_msg,
-                   product->description,
-                   uuid_s);
-  GNUNET_free (uuid_s);
-  /* create the json object for the order request */
-  if (NULL != product->preview)
+  pos = strstr (fulfillment_msg,
+                "${PRODUCT_DESCRIPTION}");
+  if (NULL != pos)
   {
-    orderReq = json_pack ("{ s:s, s:s, s:o, s:s, s:o }",
-                          "summary",
-                          product->description,
-                          "preview", product->preview,
-                          "amount", TALER_JSON_from_amount (&product->price),
-                          "fulfillment_url", fulflmntUrl
-                          ,"auto_refund", GNUNET_JSON_from_time_rel (
-                            MAX_REFUND_DELAY)
-                          );
+    /* replace ${PRODUCT_DESCRIPTION} with the real one */
+    GNUNET_asprintf (&msg,
+                     "%.*s%s%s",
+                     /* first output URL until ${PRODUCT_DESCRIPTION} */
+                     (int) (pos - fulfillment_msg),
+                     fulfillment_msg,
+                     /* replace ${PRODUCT_DESCRIPTION} with the right description */
+                     product->description,
+                     /* append rest of original URL */
+                     pos + strlen ("${PRODUCT_DESCRIPTION}"));
   }
   else
   {
-    orderReq = json_pack ("{ s:s}, s:o, s:s, s:o }",
+    msg = GNUNET_strdup (fulfillment_msg);
+  }
+
+  /* create the json object for the order request */
+  if (NULL != product->preview)
+  {
+    orderReq = json_pack (
+      "{ s:s, s:[ { s:s, s:s, s:o, s:I } ], s:o, s:s, s:o }",
+      "summary", product->description,
+      "products",
+      /* */ "description", product->description,
+      /* */ "image", product->preview,
+      /* */ "price", TALER_JSON_from_amount (
+        &product->price),
+      /* */ "quantity", (json_int_t) 1,
+      "amount", TALER_JSON_from_amount (&product->price),
+      "fulfillment_message", msg,
+      "auto_refund", GNUNET_JSON_from_time_rel (
+        MAX_REFUND_DELAY)
+      );
+  }
+  else
+  {
+    orderReq = json_pack ("{ s:s, s:o, s:s, s:o }",
                           "summary",
                           product->description,
                           "amount", TALER_JSON_from_amount (&product->price),
-                          "fulfillment_url", fulflmntUrl
-                          ,"auto_refund", GNUNET_JSON_from_time_rel (
+                          "fulfillment_message", msg,
+                          "auto_refund", GNUNET_JSON_from_time_rel (
                             MAX_REFUND_DELAY)
                           );
   }
-  GNUNET_free (fulflmntUrl);
+  GNUNET_free (msg);
   if (NULL == orderReq)
   {
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
@@ -1604,13 +1606,21 @@ launch_payment (struct Product *product)
   pa->product = product;
   pa->amount = product->price;
   /* put the order on the merchant's backend */
-  pa->po = TALER_MERCHANT_order_put (ctx,
-                                     (NULL == product->instance)
-                                     ? backend_base_url
-                                     : product->instance,
-                                     orderReq,
-                                     &proposal_cb,
-                                     pa);
+  pa->base_url = (NULL == product->instance)
+                 ? backend_base_url
+                 : product->instance;
+  pa->po = TALER_MERCHANT_orders_post2 (ctx,
+                                        pa->base_url,
+                                        orderReq,
+                                        MAX_REFUND_DELAY,
+                                        NULL, /* no payment target preference */
+                                        0,
+                                        NULL, /* no inventory */
+                                        0,
+                                        NULL, /* no locks */
+                                        false, /* no claim token */
+                                        &proposal_cb,
+                                        pa);
   json_decref (orderReq);
   if (NULL == pa->po)
   {
@@ -1647,26 +1657,27 @@ vend_success ()
  * @brief Callback to process a POST /refund request
  *
  * @param cls closure
- * @param http_status HTTP status code for this request
- * @param ec taler-specific error code
- * @param obj the response body
+ * @param hr HTTP response details
+ * @param refund_uri the refund uri offered to the wallet
+ * @param h_contract hash of the contract a Browser may need to authorize
+ *        obtaining the HTTP response.
  */
 static void
 refund_complete_cb (void *cls,
-                    unsigned int http_status,
-                    enum TALER_ErrorCode ec,
-                    const json_t *obj)
+                    const struct TALER_MERCHANT_HttpResponse *hr,
+                    const char *refund_uri,
+                    const struct GNUNET_HashCode *h_contract)
 {
   struct Refund *r = cls;
 
-  (void) obj;
-  r->rio = NULL;
-  if (MHD_HTTP_OK != http_status)
+  (void) refund_uri;
+  r->orh = NULL;
+  if (MHD_HTTP_OK != hr->http_status)
   {
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                 "Failed to grant consumer refund: %u/%d\n",
-                http_status,
-                ec);
+                hr->http_status,
+                (int) hr->ec);
   }
   GNUNET_CONTAINER_DLL_remove (refund_head,
                                refund_tail,
@@ -1679,7 +1690,7 @@ refund_complete_cb (void *cls,
  * @brief Vending failed, provide refund.
  */
 static void
-vend_failure ()
+vend_failure (void)
 {
   struct Product *p;
   struct Refund *r;
@@ -1697,16 +1708,16 @@ vend_failure ()
   mdb.cmd = &endSession;
   mdb.session_running = GNUNET_NO;
   r = GNUNET_new (struct Refund);
-  r->rio = TALER_MERCHANT_refund_increase (ctx,
-                                           (NULL == p->instance)
-                                           ? backend_base_url
-                                           : p->instance,
-                                           payment_activity->order_id,
-                                           &payment_activity->amount,
-                                           "failed to dispense product",
-                                           &refund_complete_cb,
-                                           r);
-  if (NULL == r->rio)
+  r->orh = TALER_MERCHANT_post_order_refund (ctx,
+                                             (NULL == p->instance)
+                                             ? backend_base_url
+                                             : p->instance,
+                                             payment_activity->order_id,
+                                             &payment_activity->amount,
+                                             "failed to dispense product",
+                                             &refund_complete_cb,
+                                             r);
+  if (NULL == r->orh)
   {
     GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                 "Failed to launch refund interaction with the merchant backend!\n");
@@ -2130,7 +2141,6 @@ handle_command (const char *hex,
     {
       unsigned int subcmd;
 
-      GNUNET_break (GNUNET_YES == mdb.session_running);
       if (4 > hex_len)
       {
         GNUNET_break_op (0);
@@ -2217,6 +2227,8 @@ handle_command (const char *hex,
                 return;
               }
               payment_activity = launch_payment (&products[i]);
+              if (NULL == payment_activity)
+                vend_failure ();
               return;
             }
           GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
@@ -2487,7 +2499,7 @@ read_mdb_command (void *cls)
   while (mdb.rx_off > 0)
   {
     mdb_active = GNUNET_YES;
-    /* find begining of command */
+    /* find beginning of command */
     for (cmdStartIdx = 0; cmdStartIdx < mdb.rx_off; cmdStartIdx++)
       if (VMC_CMD_START == mdb.rxBuffer[cmdStartIdx])
         break;
@@ -2550,7 +2562,7 @@ run_mdb_event_loop ()
        (NULL == mdb.last_cmd) )
   {
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-                "Begining MDB session\n");
+                "Beginning MDB session\n");
     mdb.cmd = &cmd_begin_session;
   }
   /* start write task if he doesn't exist and if there is a cmd waiting to get sent */
@@ -2650,7 +2662,7 @@ read_products (void *cls,
     GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
                                section,
                                "price",
-                               "currency missmatch");
+                               "currency mismatch");
     GNUNET_free (tmpProduct.description);
     return;
   }
@@ -2777,8 +2789,8 @@ read_products (void *cls,
                                section,
                                "number");
     GNUNET_free (tmpProduct.description);
-    GNUNET_free_non_null (tmpProduct.instance);
-    GNUNET_free_non_null (tmpProduct.preview);
+    GNUNET_free (tmpProduct.instance);
+    GNUNET_free (tmpProduct.preview);
     return;
   }
   /* append the temporary product to the existing products */
@@ -2852,7 +2864,7 @@ mdb_init ()
                          "tcsetattr");
     return GNUNET_SYSERR;
   }
-  /* if the configuration of the uart was sucessfull start the mdb write and read tasks */
+  /* if the configuration of the uart was successful start the mdb write and read tasks */
   run_mdb_event_loop ();
   return GNUNET_OK;
 }
@@ -2864,7 +2876,7 @@ mdb_init ()
  * @param cls closure
  * @param args arguments left
  * @param cfgfile config file name
- * @param cfg handle for the configuation file
+ * @param cfg handle for the configuration file
  */
 static void
 run (void *cls,
@@ -2934,6 +2946,7 @@ run (void *cls,
   }
   {
     char *auth;
+
     if (GNUNET_OK !=
         GNUNET_CONFIGURATION_get_value_string (cfg,
                                                "taler-mdb",
@@ -2953,17 +2966,16 @@ run (void *cls,
                      auth);
     GNUNET_free (auth);
   }
-  if (GNUNET_OK !=
-      GNUNET_CONFIGURATION_get_value_string (cfg,
-                                             "taler-mdb",
-                                             "fulfillment-url",
-                                             &fulfillment_url))
   {
-    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
-                               "taler-mdb",
-                               "fulfillment-url");
-    global_ret = EXIT_FAILURE;
-    return;
+    if (GNUNET_OK !=
+        GNUNET_CONFIGURATION_get_value_string (cfg,
+                                               "taler-mdb",
+                                               "essid",
+                                               &essid))
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                  "No ESSID specified, will not advertise WLAN\n");
+    }
   }
   if (GNUNET_OK !=
       GNUNET_CONFIGURATION_get_value_string (cfg,
@@ -2973,7 +2985,7 @@ run (void *cls,
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "taler-mdb",
-                               "fulfillment-msgcurrency");
+                               "fulfillment-msg");
     global_ret = EXIT_FAILURE;
     return;
   }
